#!/usr/bin/perl -w
#
# This is a program to check a network for open SMTP relays.
# It requires some configuration below, and should be run on
# a host from which relaying should be denied (eg, a local
# 192.168-network host for testing).  Make sure you set up a
# filter to send the test messages to a mailbox all of their
# own.
#
# See also listhosts.pl to print the host list up neatly from
# the contents of said mailbox.
#
# If using strobe-root-classb, read README and consider doing
# multiple runs when first using it.  If using strobe-classb
# (recommended), one run is sufficient.
#
# See the file COPYRIGHT for copyright information.
# (c) David Luyer, 1998.

use strict;
use IO::Socket;

my $reproc;

if (defined $ARGV[0] && $ARGV[0] eq 'reprocess') {
  # Reprocess entries from relay-check-hosts
  $reproc = 1;
  $ARGV[0] = "relay-check-hosts";
  open(STDERR, ">>relay-check-log");
} elsif (defined $ARGV[0]) {
  print STDERR "Only valid option is 'reprocess' to reprocess the hosts in ".
               "relay-check-hosts\nwithout a fresh scan.\n";
  exit;
} else {
  open(STDERR, ">relay-check-log");
  open(STDOUT, ">relay-check-hosts");
}

select(STDERR); $| = 1;
select(STDOUT); $| = 1;

my ($hello, $fromemail, $toemail);

# configuration for this script

### $hello : Hostname for HELO, also used in Message-Id header.
###          Usually the same as in $fromemail.
$hello = "invalid.edu";
### $fromemail : Email address used in MAIL FROM and From header.
###              This should probably be your address so you get bounces.
###              Remember - "addr\@host.domain" format.
$fromemail = "me\@invalid.net";
### $toemail : Email address used in RCPT TO.
###            This should be something which you should not relay to.
###            Remember - "addr\@host.domain" format.
$toemail = "me\@destination.edu";

# configuration for final scan task count selection

### maximum task count
my $MAXTASKS = 16;
### minimum hosts for one task
my $MINHOSTS = 16;

# configuration for scanner sub-script
### @classbs : The network(s) to scan.
@config::classbs = ( "10.0" );
### Ports : Always 25
@config::ports = ( 25 );
### Dispose : Set to 1 to delete binaries for each network/port pair, 0 to keep
$config::dispose = 1;
### Module : The module to use.  "strobe-classb" or "strobe-root-classb".
###          You should read the documentation as these are quite different.
$config::module = "strobe-classb";

my $sd;

sub WAIT {
  # Syntax: WAIT (T, X, Y, Z, A)
  # Synopsis:
  #   Execute X with an alarm at T seconds.
  #   Y is a function which returns false if the alarm has gone off.
  #   If X fails, not due to the alarm, Z is to be executed.
  #   If the alarm has gone off, A is to be executed.
  alarm($_[0]);
  &{$_[3]} if !&{$_[1]} && &{$_[2]};
  alarm(0);
  &{$_[4]} if !&{$_[2]};
}

sub skip { undef $sd; }
sub ignore { }
sub defsd { defined $sd; }

sub checkhosts {
  my ($host, $buf, $ipaddr, $dnsname);
  use POSIX ':signal_h';
  my $alarm = POSIX::SigAction->new('skip');
  my $pipe = POSIX::SigAction->new('ignore');
  my $trick = shift;

  for $host (@_) {
    $0 = "relay-check checking $host";
    $sd = IO::Socket::INET->new("$host");
    die "Error setting SIGALRM handler"
      if(!sigaction(SIGALRM, $alarm));
    die "Error reaffirming SIGPIPE handler"
      if(!sigaction(SIGPIPE, $pipe));
    print STDERR "$host:SKIPPED[Socket open failed]\n" if(!$sd);
    if($sd) {
      WAIT(30, sub {$sd->recv($buf, 1024)}, \&defsd,
	   sub {print STDERR "$host:ERROR[$!/initial banner]\n"; goto SKIP},
           sub {print STDERR "$host:TIMEOUT[initial banner]\n"; goto SKIP});
      print STDERR $host.":".$buf;
      $sd->send("HELO theoffice.net\r\n");
      WAIT(30, sub {$sd->recv($buf, 1024)}, \&defsd,
	   sub {print STDERR "$host:ERROR[$!/HELO response]\n"; goto SKIP},
           sub {print STDERR "$host:TIMEOUT[HELO response]\n"; goto SKIP});
      print STDERR $host.":".$buf;
      $sd->send("MAIL FROM: <$fromemail>\r\n");
      WAIT(30, sub {$sd->recv($buf, 1024)}, \&defsd,
	   sub {print STDERR "$host:ERROR[$!/MAIL response]\n"; goto SKIP},
           sub {print STDERR "$host:TIMEOUT[MAIL response]\n"; goto SKIP});
      print STDERR $host.":".$buf;
      if($trick) {
	$ipaddr = inet_aton(substr($host, 0, -3));
	$dnsname = gethostbyaddr($ipaddr, AF_INET);
        $sd->send("RCPT TO: <\"$toemail\"\@$dnsname>\r\n");
      } else {
        $sd->send("RCPT TO: <$toemail>\r\n");
      }
      WAIT(30, sub {$sd->recv($buf, 1024)}, \&defsd,
	   sub {print STDERR "$host:ERROR[$!/RCPT response]\n"; goto SKIP},
           sub {print STDERR "$host:TIMEOUT[RCPT response]\n"; goto SKIP});
      print STDERR $host.":".$buf;
      $sd->send("DATA\r\n");
      WAIT(30, sub {$sd->recv($buf, 1024)}, \&defsd,
	   sub {print STDERR "$host:ERROR[$!/DATA response]\n"; goto SKIP},
           sub {print STDERR "$host:TIMEOUT[DATA response]\n"; goto SKIP});
      print STDERR $host.":".$buf;
      $sd->send("From: Preventitive spam relay test <$fromemail>\r\n");
      $sd->send("Message-Id: <$host.$$.".time()."\@$hello>\r\n");
      $sd->send("Subject: relay test\r\n");
      $sd->send("\r\n");
      $sd->send("Hello via $host\r\n");
      $sd->send(".\r\n");
      WAIT(30, sub {$sd->recv($buf, 1024)}, \&defsd,
	   sub {print STDERR "$host:ERROR[$!/DATA end response]\n"; goto SKIP},
           sub {print STDERR "$host:TIMEOUT[DATA end response]\n"; goto SKIP});
      print STDERR $host.":".$buf;
      $sd->send("QUIT\r\n");
      WAIT(30, sub {$sd->recv($buf, 1024)}, \&defsd,
	   sub {print STDERR "$host:ERROR[$!/QUIT response]\n"; goto SKIP},
           sub {print STDERR "$host:TIMEOUT[QUIT response]\n"; goto SKIP});
      print STDERR $host.":".$buf;
    SKIP:
    }
  }
}

sub spawncheckhosts {
  my $count = 2 * ($#_ + 1);
  my ($tasks, $i, $from, $to);
  if($count <= $MINHOSTS) {
    checkhosts(0, @_);
    checkhosts(1, @_);
    return;
  } else {
    use integer;
    $tasks = $count/$MINHOSTS;
    no integer;
  }
  $tasks = $MAXTASKS if($tasks > $MAXTASKS);
  $SIG{CHLD} = 'IGNORE';
  for $i (1..$tasks) {
    use integer;
    $from = (($count/2) * ($i-1))/$tasks;
    $to = (($count/2) * $i)/$tasks - 1;
    no integer;
    unless(fork) {
      checkhosts(0, @_[$from..$to]);
      checkhosts(1, @_[$from..$to]);
      exit;
    }
  }
  use POSIX ":errno_h";
  # bug in POSIX module's loading of &ECHILD if @_ isn't undef'd
  undef @_;
  do {
    while (waitpid(-1, 0) != -1) {};
  } while ($! != &ECHILD);
}

if (defined $reproc) {
  $0 = "relay-check reprocessing";
  chop (@main::hosts = <>);
} else {
  $0 = "relay-check main scan";
  $main::sub = 1;
  require 'strobe-classb.pl';
}
spawncheckhosts(@main::hosts);
exit;

# Prevent perl warnings
undef $main::sub;
undef @config::classbs;
undef @config::ports;
undef $config::dispose;
undef $config::module;
